"Cartesian arrays, arrays defined over a cartesian communicator"

__all__ = [
    "cart_zeros",
    "cart_ones",
    "cart_empty",
    "cart_full",
    "cart_array",
    "get_cart_arrays",
]

from math import ceil
from dask.array import Array, zeros, ones, empty, full
from dask.array.core import normalize_chunks
from dask.distributed import wait as wait_for
from .dask_mpi import Cartcomm


def get_cart_arrays(cart, array, dims_axes=None, wait=True):
    """
    Return the list of futures of an array associated to a cartesian communicator.

    Parameters
    ----------
    cart: Cartcomm
        A cartesian communicator with dimensions equal to the number of chunks
    array: Dask Array
        A dask array distributed on the cartesian communicator
    dims_axes: tuple
        The axes associated to the dimensions of the cart
    wait: bool
        Whether to wait for the computation of the array to finish
        for checking the actual list of workers.
        Disable only if you are sure of the location of the array.
    """
    if not isinstance(cart, Cartcomm):
        cart = Cartcomm(cart)

    # As first checking if it is compatible and getting the axes
    cart_axes, arr_axes = get_axes(cart.dims, dims_axes, array.shape, array.chunks)

    array = array.persist()
    assert len(cart) == len(array.dask.values())
    if wait:
        wait_for(array.dask.values())
        workers = cart.client.who_has(array.dask.values())
        if not set(workers) == set(cart.workers):
            raise RuntimeError("Workers of cart and array do not match")

    # Creating dict of coords -> worker
    workers = {}
    for coords, worker in zip(cart.coords, cart.workers):
        workers[tuple(coords[_i] for _i in cart_axes)] = worker

    # Creating arrays list
    arrays = [None] * len(cart)
    for key, val in array.dask.items():
        assert len(key) == len(array.shape) + 1
        coords = tuple(key[_i + 1] for _i in arr_axes)
        worker = workers[coords]
        if wait:
            real_worker = cart.client.who_has(val)
            if worker not in real_worker:
                raise RuntimeError("Expected worker {worker} but got {real_worker}")
        arrays[cart.workers.index(worker)] = val

    assert None not in arrays
    return arrays


def cart_array(cart, arrays, shape, dims_axes=None, chunks=None, dtype=None):
    """
    Turns a set of future arrays (result of a distributed operation),
    associated to a cartesian communicator, into a Dask Array.

    Parameters
    ----------
    cart: Cartcomm
        A cartesian communicator with dimensions equal to the number of chunks
    arrays: tuple(futures)
        A set of future arrays associated to the cart
    dims_axes: tuple
        The axes associated to the dimensions of the cart
    shape: tuple(int)
        The shape of the array
    chunks: tuple(int)
        The chunks of the array
    dtype: tuple(int)
        The dtype of the array
    """
    if not isinstance(cart, Cartcomm):
        cart = Cartcomm(cart)

    if not len(arrays) == len(cart):
        raise ValueError("arrays and cart must have the same length")

    if chunks is None or dtype is None:
        infos = cart.client.map(lambda arr: (arr.dtype, arr.shape), arrays)
        infos = tuple(_.result() for _ in infos)

        if dtype is None:
            dtype = infos[0][0]
        if not all((dtype == dtp for (dtp, _) in infos)):
            raise TypeError(
                f"Arrays have different dtypes {[info[0] for info in infos]}"
            )

        if chunks is None:
            chunks = infos[0][1]
            if not all((chunks == chn for (_, chn) in infos)):
                # TODO: normalize chunks using shape
                raise NotImplementedError(
                    "Arrays with non-uniform chunks not supported yet"
                )

    chunks = normalize_chunks(chunks, shape, dtype=dtype)

    cart_axes, arr_axes = get_axes(cart.dims, dims_axes, shape, chunks)

    dask = {}
    for coords, array in zip(cart.coords, arrays):
        key = [0] * len(shape)
        coords = tuple(coords[_i] for _i in cart_axes)

        for _i, _j in zip(coords, arr_axes):
            key[_j] = _i

        name = arrays[0].key
        if isinstance(name, tuple):
            name = name[0]
        assert isinstance(name, str)
        key = (name,) + tuple(key)
        dask[key] = array

    return Array(dask, next(iter(dask.keys()))[0], chunks, dtype=dtype, shape=shape)


def normalize_cart_dims(dims, dims_axes=None, shape_size=None):
    "Auxiliary function that removes non-distributed dimensions from dims and dims_axes"

    cart_dims = tuple((_i, _l) for _i, _l in enumerate(dims) if _l > 1)

    if dims_axes is not None:
        if not len(dims_axes) == len(dims):
            raise ValueError(
                f"""
                Lengths of dims_axes {dims_axes} and dims {dims} do not match
                """
            )

        dims_axes = tuple(dims_axes[_i] for _i, _l in cart_dims)
        if not len(set(dims_axes)) == len(dims_axes):
            raise ValueError(f"Repeated values in dims_axes {dims_axes}")
        if not set(dims_axes) <= set(range(shape_size)):
            raise ValueError(
                f"Values in dims_axes {dims_axes} out of shape range {shape_size}"
            )

    return cart_dims, dims_axes


def get_axes(dims, dims_axes, shape, chunks):
    "Auxiliary function that matches the dims_axes with the chunks"

    cart_dims, dims_axes = normalize_cart_dims(dims, dims_axes, len(shape))
    cdims = tuple(_l for _i, _l in cart_dims)

    # Extracting the chunked axes and cart dims
    arr_dims = tuple((_i, _l) for _i, _l in enumerate(map(len, chunks)) if _l > 1)
    adims = tuple(_l for _i, _l in arr_dims)

    if sorted(adims) != sorted(cdims):
        raise ValueError(
            f"""
            Number of chunks {adims} different from cart dims {cdims}.
            """
        )

    # Reording cart.dims with dims_axes
    if dims_axes is not None:
        aaxes = tuple(_i for _i, _l in arr_dims)
        if not sorted(dims_axes) == sorted(aaxes):
            raise ValueError(
                f"""
                Cart axes {sorted(aaxes)} different from chunked axes {sorted(dims_axes)}.
                """
            )
        cart_dims = tuple(cart_dims[dims_axes.index(_i)] for _i in aaxes)
        cdims = tuple(_l for _i, _l in cart_dims)

    if adims != cdims:
        raise ValueError(
            f"""
            Number of chunks {adims} and cart dims {cdims} do not match.
            You can use dims_axes to reorder them.
            """
        )
    return tuple(_i for _i, _l in cart_dims), tuple(_i for _i, _l in arr_dims)


def array_wrapper(mth):
    """
    Cartesian variant of %s (documentation follows).
    Assigns to the workers of a cartesian communicator chunks of the array.

    The first argument `cart` should be a cartesian communicator
    with dimensions equal to the number of chunks of the array.

    Additional option is `dims_axes` that can be used to reorder
    the dimensions of the communicator to match the chunked axes.

    --------- END ----------
    """

    def wrapper(cart, shape, dims_axes=None, chunks=None, **kwargs):
        if not isinstance(cart, Cartcomm):
            cart = Cartcomm(cart)

        if chunks is None:
            chunks = list(shape)
            loop_over = enumerate(cart.dims)
            if dims_axes:
                cart_dims, cart_axes = normalize_cart_dims(
                    cart.dims, dims_axes, len(shape)
                )
                cdims = tuple(_l for _i, _l in cart_dims)
                loop_over = zip(cart_axes, cdims)
            for _i, _l in loop_over:
                if _l > shape[_i]:
                    raise ValueError(f"Cannot chunk shape on axis {_i} in {_l} pieces")
                chunks[_i] = ceil(chunks[_i] / _l)

        arr = mth(shape, chunks=chunks, **kwargs)

        cart_axes, arr_axes = get_axes(cart.dims, dims_axes, arr.shape, arr.chunks)

        # Creating dict of coords -> worker
        workers = {}
        for coords, worker in zip(cart.coords, cart.workers):
            workers[tuple(coords[_i] for _i in cart_axes)] = worker

        # Creating dask dict
        dask = {}
        for key, val in arr.dask.items():
            coords = tuple(key[_i + 1] for _i in arr_axes)
            dask[key] = cart.client.submit(*val, workers=[workers[coords]], pure=False)

        return Array(
            dask,
            next(iter(dask.keys()))[0],
            arr.chunks,
            dtype=arr.dtype,
            meta=arr._meta,
            shape=arr.shape,
        )

    wrapper.__name__ = "cart_" + mth.__name__
    wrapper.__doc__ = array_wrapper.__doc__ % mth.__name__ + mth.__doc__
    return wrapper


cart_zeros = array_wrapper(zeros)
cart_ones = array_wrapper(ones)
cart_empty = array_wrapper(empty)
cart_full = array_wrapper(full)



"Loading and using MPI via cppyy"

__all__ = [
    "lib",
    "default_comm",
    "initialized",
    "finalized",
    "initialize",
    "finalize",
]

from ctypes import c_int
from cppyy import nullptr
from lyncs_cppyy import Lib
from .config import MPI_INCLUDE_DIRS, MPI_LIBRARIES

lib = Lib(
    include=MPI_INCLUDE_DIRS.split(";"),
    header="mpi.h",
    library=MPI_LIBRARIES.split(";"),
    c_include=False,
    check="MPI_Init",
)


COMM = None


def default_comm():
    "Returns the default communicator to be used (MPI_COMM_WORLD by default)"
    # pylint: disable=import-outside-toplevel,no-name-in-module,redefined-outer-name,global-statement
    global COMM
    if not COMM:
        from mpi4py.MPI import COMM_WORLD as COMM

    return COMM


def initialized():
    "Whether MPI has been initialized"
    val = c_int(0)
    lib.MPI_Initialized(val)
    return bool(val)


def finalized():
    "Whether MPI has been finalized"
    val = c_int(0)
    lib.MPI_Finalized(val)
    return bool(val)


def initialize():
    "Initializes MPI"
    assert not initialized() and not finalized()
    lib.MPI_Init(nullptr, nullptr)


def finalize():
    "Finalizes MPI"
    assert initialized() and not finalized()
    lib.MPI_Finalize()



from pytest import raises
from lyncs_mpi import (
    Client,
    default_client,
    cart_zeros,
    cart_ones,
    get_cart_arrays,
    cart_array,
)
from dask.distributed import wait


def test_client():
    client = Client(num_workers=1, launch=True)
    assert client._server is not None
    assert default_client() is client
    assert len(client.workers) == 1

    client.close_server()
    assert client._server is None


def test_comm():
    client = Client(num_workers=4, launch=True)
    comm = client.comm
    assert len(comm) == len(client.workers)
    assert len(comm) == comm.size
    assert set(comm.ranks) == set(range(4))
    assert set(comm.workers) == set(client.workers)
    assert comm[0] == comm[comm.ranks_worker[0]]

    with raises(KeyError):
        comm[5]

    cart = comm.create_cart([2, 2], periods=True)
    assert cart.dims == (2, 2)
    assert all(cart.periods)
    assert cart[0] == cart[cart.ranks_coords[0]]


def test_array():
    client = Client(num_workers=4, launch=True)
    cart = client.comm.create_cart([2, 2], periods=True)

    arr = cart_zeros(cart, (4, 4), chunks=(2, 2))
    assert arr.sum() == 0

    arrs = get_cart_arrays(cart, arr)
    assert len(arrs) == len(cart)

    arr = cart_array(cart, arrs, (4, 4))
    assert arr.sum() == 0
    assert tuple(arr.dask.keys()) == tuple(val.key for val in arr.dask.values())

    arr = cart_zeros(tuple(cart), (4, 4), chunks=(2, 2))
    assert arr.sum() == 0

    arrs = get_cart_arrays(tuple(cart), arr)
    assert len(arrs) == len(cart)

    arr = cart_array(tuple(cart), arrs, (4, 4))
    assert arr.sum() == 0
    assert tuple(arr.dask.keys()) == tuple(val.key for val in arr.dask.values())

    arr = cart_ones(cart, (2, 4, 4), chunks=(2, 2, 2))
    assert arr.sum() == 32

    arrs = get_cart_arrays(cart, arr)
    assert len(arrs) == len(cart)

    arr = cart_array(cart, arrs, (2, 4, 4))
    assert arr.sum() == 32
    assert tuple(arr.dask.keys()) == tuple(val.key for val in arr.dask.values())

    arr = cart_ones(cart, (4, 4, 2))
    assert arr.sum() == 32

    arr = cart_ones(cart, (2, 4, 4), dims_axes=(1, 2))
    assert arr.sum() == 32

    arrs = get_cart_arrays(cart, arr, dims_axes=(1, 2))
    assert len(arrs) == len(cart)

    arr = cart_array(cart, arrs, (2, 4, 4), dims_axes=(1, 2), chunks=(2, 2, 2))
    assert arr.sum() == 32
    assert tuple(arr.dask.keys()) == tuple(val.key for val in arr.dask.values())

    arr = cart_ones(cart, (4, 2, 4), dims_axes=(2, 0))
    assert arr.sum() == 32

    arrs = get_cart_arrays(cart, arr, dims_axes=(2, 0))
    assert len(arrs) == len(cart)

    arr = cart_array(cart, arrs, (4, 2, 4), dims_axes=(2, 0))
    assert arr.sum() == 32
    assert tuple(arr.dask.keys()) == tuple(val.key for val in arr.dask.values())

    # Testing errors
    with raises(ValueError):
        arr = cart_zeros(cart, (6, 4), chunks=(2, 2))

    with raises(ValueError):
        arr = cart_zeros(cart, (4, 4, 4), chunks=(2, 2, 2))

    with raises(ValueError):
        arr = cart_zeros(cart, (4, 4), chunks=(2, 2), dims_axes=(0, 1, 2))

    with raises(ValueError):
        arr = cart_zeros(cart, (2, 4, 4), chunks=(2, 2, 2), dims_axes=(0, 1))

    with raises(ValueError):
        arr = cart_zeros(cart, (1, 1))

    with raises(ValueError):
        arr = cart_zeros(cart, (4, 4), dims_axes=(0, 0))

    with raises(ValueError):
        arr = cart_zeros(cart, (4, 4), dims_axes=(3, 4))

    with raises(ValueError):
        arr = cart_array(cart, arrs[:-1], (4, 2, 4))

    # Testing a weird cart
    cart = client.comm.create_cart([1, 2, 1, 1, 2, 1], periods=True)
    arr = cart_ones(cart, (4, 2, 4), chunks=(2, 2, 2))
    assert arr.sum() == 32

    arr = cart_ones(
        cart, (1, 4, 2, 4, 1), chunks=(1, 2, 2, 2, 1), dims_axes=(-1, 3, -1, -1, 1, -1)
    )
    assert arr.sum() == 32

    # Testing a cart with different dims
    client = Client(num_workers=6, launch=True)
    cart = client.comm.create_cart([3, 2], periods=True)
    arr = cart_ones(cart, (6, 4), chunks=(2, 2))
    assert arr.sum() == 24

    arr = cart_ones(cart, (4, 6), chunks=(2, 2), dims_axes=(1, 0))
    assert arr.sum() == 24

    arrs = get_cart_arrays(cart, arr, dims_axes=(1, 0))
    assert len(arrs) == len(cart)

    arr = cart_array(cart, arrs, (4, 6), dims_axes=(1, 0))
    assert arr.sum() == 24
    assert tuple(arr.dask.keys()) == tuple(val.key for val in arr.dask.values())

    with raises(ValueError):
        arr = cart_ones(cart, (4, 6), chunks=(2, 2))



from lyncs_mpi import initialized, initialize, finalize, finalized


def test_init():
    if not initialized():
        initialize()
    from mpi4py import MPI

    assert initialized() == True
    assert MPI.Is_initialized() == True
    finalize()
    assert finalized() == True
    assert MPI.Is_finalized() == True



"""
Utils for interfacing dask and MPI.
Most of the content of this file is explained in the [notebook](notebooks/Dask-mpi.ipynb).
"""

__all__ = [
    "default_client",
    "Client",
]

import os
import time
import shutil
import signal
import atexit
import tempfile
import multiprocessing
from functools import wraps

import sys
from packaging import version
import sh

# The following can be deleted when sh==1.13.2 has been released
if version.parse(sh.__version__) < version.parse("1.13.2"):

    sys.meta_path.pop(0)

    def find_spec(self, fullname, path=None, target=None):
        """ find_module() is deprecated since Python 3.4 in favor of find_spec() """

        from importlib.machinery import ModuleSpec

        found = self.find_module(fullname, path)
        return ModuleSpec(fullname, found) if found is not None else None

    sh._SelfWrapper__self_module.ModuleImporterFromVariables.find_spec = find_spec
    sh._SelfWrapper__self_module.register_importer()


from dask_mpi import initialize
from dask.distributed import Client as _Client
from dask.distributed import default_client as _default_client
from .lib import default_comm


@wraps(_default_client)
def default_client():
    "Returns the default client"
    client = _default_client()
    assert isinstance(client, Client), "No MPI client found"
    return client


class Client(_Client):
    "Wrapper to dask.distributed.Client"

    def __init__(self, num_workers=None, threads_per_worker=1, launch=None):
        """
        Returns a Client connected to a cluster of `num_workers` workers.
        """
        self._server = None

        if launch is None:
            launch = default_comm().size == 1

        # pylint: disable=import-outside-toplevel,
        if not launch:
            # Then the script has been submitted in parallel with mpirun
            num_workers = num_workers or default_comm().size - 1
            assert (
                default_comm().size == num_workers + 1
            ), """
            Error: (num_workers + 1) processes required.
            The script has not been submitted on enough processes.
            Got %d processes instead of %d.
            """ % (
                default_comm().size,
                num_workers + 1,
            )

            initialize(nthreads=threads_per_worker, nanny=False)

            _Client.__init__(self)

        else:
            num_workers = num_workers or (multiprocessing.cpu_count() + 1)

            # Since dask-mpi produces several file we create a temporary directory
            self._dir = tempfile.mkdtemp()
            self._out = self._dir + "/log.out"
            self._err = self._dir + "/log.err"

            # The command runs in the background (_bg=True)
            # and the stdout(err) is stored in self._out(err)
            pwd = os.getcwd()
            sh.cd(self._dir)
            self._server = sh.mpirun(
                "-n",
                num_workers + 1,
                "dask-mpi",
                "--no-nanny",
                "--nthreads",
                threads_per_worker,
                "--scheduler-file",
                "scheduler.json",
                _bg=True,
                _out=self._out,
                _err=self._err,
            )
            sh.cd(pwd)

            atexit.register(self.close_server)

            _Client.__init__(self, scheduler_file=self._dir + "/scheduler.json")

        # Waiting for all the workers to connect
        def handler(signum, frame):
            if self._server is not None:
                self.close_server()
            raise RuntimeError(
                "Couldn't connect to %d processes. Got %d workers."
                % (num_workers, len(self.workers))
            )

        signal.signal(signal.SIGALRM, handler)
        signal.alarm(5)

        while len(self.workers) != num_workers:
            time.sleep(0.001)

        signal.alarm(0)

        self.ranks = {key: val["name"] for key, val in self.workers.items()}
        self._comm = self.create_comm()

    @property
    def comm(self):
        "Returns the global communicator of the clients"
        return self._comm

    @property
    def workers(self):
        "Returns the list of workers."
        return self.scheduler_info()["workers"]

    def close_server(self):
        """
        Closes the running server
        """
        assert self._server is not None
        self.shutdown()
        self.close()
        self._server.wait()
        shutil.rmtree(self._dir)
        self._server = None
        atexit.unregister(self.close_server)

    def __del__(self):
        """
        In case of server started, closes the server
        """
        if self._server is not None:
            self.close_server()

        if hasattr(_Client, "__del__"):
            _Client.__del__(self)

    def who_has(self, *args, overload=True, **kwargs):
        """
        Overloading of distributed.Client who_has.
        Checks that only one worker owns the futures and returns the list of workers.

        Parameters
        ----------
        overload: bool, default true
            If false the original who_has is used
        """
        if overload:
            _workers = list(_Client.who_has(self, *args, **kwargs).values())
            workers = [w[0] for w in _workers if len(w) == 1]
            assert len(workers) == len(
                _workers
            ), "More than one process has the same reference"
            return workers

        return _Client.who_has(self, *args, **kwargs)

    def select_workers(
        self, num_workers=None, workers=None, exclude=None, resources=None
    ):
        """
        Selects `num_workers` from the one available.

        Parameters
        ----------
        workers: list, default all
          List of workers to choose from.
        exclude: list, default none
            List of workers to exclude from the total.
        resources: dict, default none
            Defines the resources the workers should have.
        """

        if not workers:
            workers = list(self.ranks.keys())

        workers = set(workers)
        workers = workers.intersection(self.ranks.keys())

        if exclude:
            if isinstance(exclude, str):
                exclude = [exclude]
            workers = workers.difference(exclude)

        if resources:
            # TODO select accordingly requested resources
            raise NotImplementedError("Resources not implemented.")

        if not num_workers:
            num_workers = len(workers)

        if num_workers > len(workers):
            raise RuntimeError("Available workers are less than required")

        # TODO implement some rules to choose wisely n workers
        # e.g. workers less busy, close to each other, etc
        selected = list(workers)[:num_workers]

        return selected

    def create_comm(self, actor=False, **kwargs):
        """
        Return a MPI communicator involving workers available by the client.

        Parameters
        ----------
        actor: bool, default True
            Wether the returned communicator should be a dask actor.
        **kwargs: params
            Following list of parameters for the function select_workers.
        """

        workers = self.select_workers(**kwargs)
        ranks = [[self.ranks[w] for w in workers]] * len(workers)
        ranks = self.scatter(ranks, workers=workers, hash=False, broadcast=False)

        # Checking the distribution of the group
        _workers = self.who_has(ranks)
        assert set(workers) == set(
            _workers
        ), """
        Error: Something wrong with scatter. Not all the workers got a piece.
        Expected workers = %s
        Got workers = %s
        """ % (
            workers,
            _workers,
        )

        def _create_comm(ranks):
            comm = default_comm()
            return comm.Create_group(comm.group.Incl(ranks))

        return Comm(self.map(_create_comm, ranks, actor=actor))


class Comm:
    "MPI communicator"

    __slots__ = [
        "_comms",
        "_ranks",
        "_workers",
    ]

    def __init__(self, comms):
        self._comms = tuple(comms)
        self._ranks = self.client.map(lambda comm: comm.rank, self)
        self._ranks = tuple(_.result() for _ in self._ranks)
        self._workers = tuple(map(next, map(iter, map(self.client.who_has, self))))
        assert len(set(self.workers)) == len(self), "Not unique set of workers"

    @property
    def client(self):
        "Client of the communicator"
        return self._comms[0].client

    @property
    def size(self):
        "Size of the communicator"
        return len(self)

    @property
    def ranks(self):
        "Ranks of the communicator with respective worker"
        return self._ranks

    @property
    def workers(self):
        "Workers of the communicator with respective rank"
        return self._workers

    @property
    def ranks_worker(self):
        "Mapping between ranks and workers"
        return dict(zip(self._ranks, self._workers))

    def create_cart(self, dims, periods=None, reorder=False):
        """
        Makes a new communicator to which topology information has been attached

        Parameters
        ----------
        dims: list
            integer array specifying the number of processes in each dimension
        periods: boolean [list]
            logical [array] specifying whether the grid is periodic (True) or not (False)
        reorder: boolean
            ranking may be reordered (True) or not (False)
        """
        return Cartcomm(
            self.client.map(
                lambda comm: comm.Create_cart(dims, periods=periods, reorder=reorder),
                self,
            )
        )

    def __getitem__(self, key):
        if isinstance(key, int) and key in self.ranks:
            return self._comms[self.ranks.index(key)]
        if isinstance(key, str) and key in self.workers:
            return self._comms[self.workers.index(key)]
        raise KeyError(f"{key} is neither a rank or a worker of {self}")

    def __len__(self):
        return len(self._comms)

    def __iter__(self):
        return iter(self._comms)


class Cartcomm(Comm):
    "Cartesian communicator"

    __slots__ = [
        "_dims",
        "_periods",
        "_coords",
    ]

    def __init__(self, comms):
        super().__init__(comms)
        topos = self.client.map(lambda comm: comm.Get_topo(), self)
        topos = tuple(_.result() for _ in topos)
        self._dims = tuple(topos[0][0])
        self._periods = tuple(topos[0][1])
        self._periods = tuple(bool(_) for _ in self._periods)
        self._coords = tuple(tuple(topo[2]) for topo in topos)

    @property
    def dims(self):
        "Dimensions of the cartesian communicator"
        return self._dims

    @property
    def periods(self):
        "Periodicity of the cartesian communicator"
        return self._periods

    @property
    def coords(self):
        "Coordinates of the cartesian communicator"
        return self._coords

    @property
    def ranks_coords(self):
        "Coordinates of the ranks of the cartesian communicator"
        return dict(zip(self.ranks, self.coords))

    def __getitem__(self, key):
        if isinstance(key, tuple) and key in self.coords:
            return self._comms[self.coords.index(key)]
        return super().__getitem__(key)



"""
This file is automitically generated during the setup.
Please do not edit config.py, but edit config.py.in in the root directory.
"""

MPI_INCLUDE_DIRS = "/home/Simone/anaconda3/include"
MPI_LINK_FLAGS = "-Wl,-rpath,/home/Simone/anaconda3/lib"
MPI_LIBRARIES = "/home/Simone/anaconda3/lib/libmpi.so"



from lyncs_setuptools import setup, CMakeExtension

requirements = [
    "mpi4py",
    "lyncs-cppyy",
    "dask",
    "distributed",
    "dask[array]",
    "dask[distributed]",
    "dask-mpi",
    "sh",
]


setup(
    "lyncs_mpi",
    exclude=["*.config"],
    ext_modules=[CMakeExtension("lyncs_mpi.lib", ".")],
    data_files=[(".", ["config.py.in"])],
    install_requires=requirements,
    extras_require={
        "test": [
            "pytest",
            "pytest-cov",
        ]
    },
    keywords=[
        "Lyncs",
        "MPI",
        "mpi4py",
        "dask-mpi",
    ],
)
